////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Project   : Fangorn Wizards Lab Exstension Library v1.35
//  * Unit Name : FWOpenDirectory
//  * Purpose   :       .
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2011.
//  * Version   : 1.11
//  ****************************************************************************
//

//  :
// 22  2003 -   ShlOBJ,
// -    
//   Delphi

// 29  2005 -    
//    / 
// - DisableOkOnStart,   OnChangeFolder

// 6  2006 -       aka Piter,
//     .     
// pszDisplayName  _browseinfoA -    MSDN :)

// 29  2006 -      Execute
//      
//  .
//    OnChangeFolder   
//     ""   .
//   ShowScreenCenter    .

// 7  2006 -   EditReadOnly
//     EDIT    

// 13  2010 -   IncludeTrailingPathDelimiter 
//        OnChangeFolder

// 25  2011 -    Delphi 2010
//       .

// 16  2011 -       Delphi 2010

{$RANGECHECKS OFF}


unit FWOpenDirectory;

interface

uses
  Windows, Messages, Classes, Controls;

const
  SID_IShellFolder       = '{000214E6-0000-0000-C000-000000000046}';
  BFFM_INITIALIZED       = 1;
  BFFM_SELCHANGED        = 2;
  BFFM_VALIDATEFAILED    = 3;
  BFFM_SETSTATUSTEXT     = WM_USER + 100;
  BFFM_ENABLEOK          = WM_USER + 101;
  BFFM_SETSELECTIONA     = WM_USER + 102;
  BIF_RETURNONLYFSDIRS   = $0001;
  BIF_DONTGOBELOWDOMAIN  = $0002;  
  BIF_STATUSTEXT         = $0004;
  BIF_RETURNFSANCESTORS  = $0008;
  BIF_EDITBOX            = $0010;
  BIF_VALIDATE           = $0020;  
  BIF_NEWDIALOGSTYLE     = $0040;
  BIF_BROWSEINCLUDEURLS  = $0080;
  BIF_BROWSEFORCOMPUTER  = $1000;  
  BIF_BROWSEFORPRINTER   = $2000;  
  BIF_BROWSEINCLUDEFILES = $4000;
  BIF_SHAREABLE          = $8000;
  BIF_USENEWUI = BIF_NEWDIALOGSTYLE or BIF_EDITBOX;

type
  PSHItemID = ^TSHItemID;
  _SHITEMID = record
    cb: Word;
    abID: array[0..0] of Byte;
  end;
  TSHItemID = _SHITEMID;

  PItemIDList = ^TItemIDList;
  _ITEMIDLIST = record
     mkid: TSHItemID;
   end;
  TItemIDList = _ITEMIDLIST;

  BFFCALLBACK = function(Wnd: HWND; uMsg: UINT; lParam, lpData: LPARAM): Integer stdcall;
  TFNBFFCallBack = type BFFCALLBACK;

  PBrowseInfo = ^TBrowseInfo;
  _browseinfoA = record
    hwndOwner: HWND;
    pidlRoot: PItemIDList;
    pszDisplayName: PAnsiChar;
    lpszTitle: PAnsiChar;
    ulFlags: UINT;
    lpfn: TFNBFFCallBack;
    lParam: LPARAM;
    iImage: Integer;
  end;
  TBrowseInfoA = _browseinfoA;
  TBrowseInfo = TBrowseInfoA;

  TFWChangeFolderEvent = procedure(Sender: TObject; Path: string;
    var EnableButton: Boolean) of object;
  TDoChangeFolder = procedure(Path: string; var EnableButton: Boolean) of object;

  TODCallbackBaram = record
    StartDir: Integer;
    Title: Integer;
    StatusText: Integer;
    ChangeFolder: TDoChangeFolder;
    Centered: Boolean;
    EditRO: Boolean;
  end;
  PODCallbackBaram = ^TODCallbackBaram;

  TFWOpenDirStyle = (odsReturnOnlyFSDirs, odsDontGoBelowDomain, odsStatusText,
    odsReturnFSAncestors, odsEditBox, odsValidate, odsNewDialogStyle,
    odsBrowseIncludeUrls, odsBrowseForComputer, odsBrowseForPrinter,
    odsBrowseIncludeFiles, odsShareable, odsUseNewUI);
  TFWOpenDirStyles = set of TFWOpenDirStyle;

  IShellFolder = interface(IUnknown) [SID_IShellFolder]
    function ParseDisplayName(hwndOwner: HWND;
      pbcReserved: Pointer; lpszDisplayName: PWideChar; out pchEaten: ULONG;
      out ppidl: PItemIDList; var dwAttributes: ULONG): HResult; stdcall;
  end;

  //       
  TFWOpenDirectory = class(TComponent)
  private
    FAbout: string;
    FTitle: string;
    FStatusText: string;
    FDirName: string;
    FStartDir: string;
    FRootDir: WideString;
    FStyle: TFWOpenDirStyles;
    FStyleArray: array [0..12] of Integer;
    FHandle: THandle;
    FShowScreenCenter: Boolean;
    FEditReadOnly: Boolean;
    FOnChangeFolder: TFWChangeFolderEvent;
  protected
    procedure Init; virtual;
    procedure DoChangeFolder(Path: string; var EnableButton: Boolean); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    function Execute(OwnerHandle: THandle = INVALID_HANDLE_VALUE): Boolean;
    property DirName: string read FDirName;
  published
    property About: string read FAbout write FAbout;
    property StatusText: string read FStatusText write FStatusText nodefault;
    property RootDir: WideString read FRootDir write FRootDir;
    property StartDir: string read FStartDir write FStartDir nodefault;
    property Style: TFWOpenDirStyles read FStyle write FStyle default [odsUseNewUI];
    property Title: string read FTitle write FTitle nodefault;
    property ShowScreenCenter: Boolean read FShowScreenCenter write FShowScreenCenter default True;
    property EditReadOnly: Boolean read FEditReadOnly write FEditReadOnly default False;
    property OnChangeFolder: TFWChangeFolderEvent read FOnChangeFolder write FOnChangeFolder;
  end;

  function SHBrowseForFolder(var lpbi: TBrowseInfo): PItemIDList; stdcall;
  function SHGetPathFromIDList(pidl: PItemIDList; pszPath: PAnsiChar): BOOL; stdcall;
  function SHGetDesktopFolder(var ppshf: IShellFolder): HResult; stdcall;
  function CoInitialize(pvReserved: Pointer): HResult; stdcall;
  procedure CoUninitialize; stdcall;
  function SHBrowseForFolder; external 'shell32.dll' name 'SHBrowseForFolderA';
  function SHGetPathFromIDList; external 'shell32.dll' name 'SHGetPathFromIDListA';
  function SHGetDesktopFolder; external 'shell32.dll' name 'SHGetDesktopFolder';
  function CoInitialize; external 'ole32.dll' name 'CoInitialize';
  procedure CoUninitialize; external 'ole32.dll' name 'CoUninitialize';  

implementation

uses ComObj, ActiveX, SysUtils;

{ TFWOpenDirectory }

constructor TFWOpenDirectory.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner = nil then
    FHandle := GetDesktopWindow
  else
    FHandle := TWinControl(AOwner).Handle;
  FStyle := [odsUseNewUI];
  FStartDir := 'c:\';
  FTitle := 'Browse For Folder';
  FStatusText := 'Specify the directory:';
  FShowScreenCenter := True;
  FEditReadOnly := False;
  Init;
end;

procedure TFWOpenDirectory.DoChangeFolder(Path: string;
  var EnableButton: Boolean);
begin
  if Assigned(FOnChangeFolder) then
    FOnChangeFolder(Self, Path, EnableButton);
end;

var
  OldProc: DWORD;

//  
function TFWOpenDirectory.Execute(
  OwnerHandle: THandle = INVALID_HANDLE_VALUE): Boolean;

  //      
  //    ?
  //   BIF_USENEWUI   BFFM_INITIALIZED
  //  GetWindowRect     ,
  //      .
  // -     
  function ExchangeDlgProc(hwndDlg: HWND; uMsg: UINT;
    WParam: WPARAM; LParam: LPARAM): LRESULT; stdcall;
  var
    ScrWidth, ScrHeight, CurWidth,
    CurHeight, Left, Top: Cardinal;
    ClientRect: TRect;
  begin
    if uMsg = WM_SHOWWINDOW then
    begin
      GetWindowRect(hwndDlg, ClientRect);
      CurWidth := ClientRect.Right - ClientRect.Left;
      CurHeight := ClientRect.Bottom - ClientRect.Top;
      ScrWidth := GetSystemMetrics(SM_CXSCREEN);
      ScrHeight := GetSystemMetrics(SM_CYSCREEN);
      Left := (ScrWidth - CurWidth) div 2;
      Top := (ScrHeight - CurHeight) div 2;
      SetWindowPos(hwndDlg, 0, Left, Top, 0, 0, SWP_NOSIZE or SWP_NOZORDER);
    end;
    Result := CallWindowProc(Pointer(OldProc),
      hwndDlg, uMsg, WParam, LParam);
  end;

  // Callback  
  //         
  function BrowseCallbackProc(hwnd: HWND; uMsg: UINT;
    lParam, lpData: LPARAM): Integer; stdcall;
  const
    ErrorPath = 'The name entered was not valid.';
  var
    EnableButton: Boolean;
    Path: array[0..MAX_PATH - 1] of AnsiChar;
    hEdit: THandle;
    strPath: string;
    aStartDir: AnsiString;
  begin
    Result := 0;
    case uMsg of
      //    
      BFFM_INITIALIZED:
      begin
        //  
        SendMessage(hwnd, WM_SETTEXT, 0, PODCallbackBaram(lpData)^.Title);
        //   
        SendMessage(hwnd, BFFM_SETSTATUSTEXT, 0, PODCallbackBaram(lpData)^.StatusText);
        //    
        aStartDir := AnsiString(PChar(PODCallbackBaram(lpData)^.StartDir));
        SendMessage(hwnd, BFFM_SETSELECTIONA, 1, Integer(@aStartDir[1]));
        //     
        EnableButton := True;
        PODCallbackBaram(lpData)^.ChangeFolder(
          PChar(PODCallbackBaram(lpData)^.StartDir), EnableButton);
        SendMessage(hwnd, BFFM_ENABLEOK, 0, Integer(EnableButton));

        //    
        hEdit := FindWindowEx(hwnd, 0, 'EDIT', nil);

        //     
        if PODCallbackBaram(lpData)^.EditRO then
          if hEdit <> 0 then
            EnableWindow(hEdit, False);

        //   -  
        if PODCallbackBaram(lpData)^.Centered then
          //       
          //    DWL_USER , ..  
          //   PBFSF    
          OldProc := SetWindowLong(hwnd, DWL_DLGPROC,
            DWORD(@ExchangeDlgProc));
      end;

      //   
      BFFM_SELCHANGED:
      begin
        //   OK
        EnableButton := True;
        if lParam <> 0 then
        begin
          Path[0] := #0;
          SHGetPathFromIDList(PItemIDList(lParam), Path);
          // Rouse_ 13.04.2010
          //   
          strPath := IncludeTrailingPathDelimiter(string(PAnsiChar(@Path[0])));
          PODCallbackBaram(lpData)^.ChangeFolder(strPath, EnableButton);
          SendMessage(hwnd, BFFM_ENABLEOK, 0, Integer(EnableButton));
          SendMessage(hwnd, BFFM_SETSTATUSTEXT, 0, Integer(@strPath[1]));
        end;
      end;
      
      //   
      BFFM_VALIDATEFAILED:
      begin
        //   
        SendMessage(hwnd, BFFM_ENABLEOK, 0, 0);
        //    odsStatusText -     
        SendMessage(hwnd, BFFM_SETSTATUSTEXT, 0, Integer(@ErrorPath[1]));
      end;
    end;
  end;

var
  lpItemID, lpRootID: PItemIDList;
  BrowseInfo: TBrowseInfo;
  Desktop: IShellFolder;
  TempPath: array[0..MAX_PATH - 1] of AnsiChar;
  pchEaten: Cardinal;
  dwAttributes: LongWord;
  I, Flags: Integer;
  CallbackParam: TODCallbackBaram;
  ShellMalloc: IMalloc;
  TmpHandle: THandle;
begin
  if OwnerHandle = INVALID_HANDLE_VALUE then
    TmpHandle := FHandle
  else
    TmpHandle := OwnerHandle;
    
  if (CoGetMalloc(1, ShellMalloc) <> S_OK) or (ShellMalloc = nil) then
    raise EComponentError.Create('CoGetMalloc failed.');
  FDirName := '';
  CoInitializeEx(nil, COINIT_APARTMENTTHREADED);
  try
    //       ( )
    Flags := 0;
    for I:= 0 to 12 do
      if (TFWOpenDirStyle(I) in FStyle) then
        Flags:= Flags or FStyleArray[I];
    //  Root    desktop folder    PIDL
    lpRootID := nil;
    if FRootDir <> '' then
    begin
      SHGetDesktopFolder(Desktop);
      dwAttributes := 0;
      Desktop.ParseDisplayName(TmpHandle, nil, PWideChar(FRootDir),
        pchEaten, lpRootID, dwAttributes);
    end;
    try
      //  
      FillChar(BrowseInfo, SizeOf(TBrowseInfo), #0);
      with BrowseInfo do
      begin
        hwndOwner := TmpHandle;
        pidlRoot := lpRootID;
        lpszTitle := PAnsiChar(AnsiString(FStatusText));
        lpfn := @BrowseCallbackProc;
        //      -    CallbackProc
        CallbackParam.StartDir := Integer(@FStartDir[1]);
        //      CallbackProc
        CallbackParam.Title := Integer(@FTitle[1]);
        CallbackParam.StatusText := Integer(@FStatusText[1]);
        CallbackParam.ChangeFolder := DoChangeFolder;
        CallbackParam.Centered := FShowScreenCenter;
        CallbackParam.EditRO := FEditReadOnly;
        lParam := Integer(@CallbackParam);
        ulFlags := Flags;
      end;
      //  
      lpItemID := SHBrowseForFolder(BrowseInfo);
      //  
      if Assigned(lpItemId) then
      try
        TempPath[0] := #0;
        Result := SHGetPathFromIDList(lpItemID, TempPath);
        // Rouse_ 13.04.2010
        //   
        FDirName := IncludeTrailingPathDelimiter(string(PAnsiChar(@TempPath[0])));
      finally
        ShellMalloc.Free(lpItemID);
      end
      else
        Result := False;
    finally
      if lpRootID <> nil then
        ShellMalloc.Free(lpRootID);
    end;
  finally
    CoUninitialize;
  end;
end;

//       
procedure TFWOpenDirectory.Init;
begin
  FStyleArray[0] := BIF_RETURNONLYFSDIRS;
  FStyleArray[1] := BIF_DONTGOBELOWDOMAIN;
  FStyleArray[2] := BIF_STATUSTEXT;
  FStyleArray[3] := BIF_RETURNFSANCESTORS;
  FStyleArray[4] := BIF_EDITBOX;
  FStyleArray[5] := BIF_VALIDATE;
  FStyleArray[6] := BIF_NEWDIALOGSTYLE;
  FStyleArray[7] := BIF_BROWSEINCLUDEURLS;
  FStyleArray[8] := BIF_BROWSEFORCOMPUTER;
  FStyleArray[9] := BIF_BROWSEFORPRINTER;
  FStyleArray[10] := BIF_BROWSEINCLUDEFILES ;
  FStyleArray[11] := BIF_SHAREABLE;
  FStyleArray[12] := BIF_USENEWUI;
end;

end.


